# -*- coding: utf-8 -*-
"""
Created on Mon May  9 18:47:48 2022

@author: tobias.haug@u.nus.edu

TFIM ground state and kink-W-State.
Calculates Clifford transformation from W-state to kink-W-State


"""




import qutip as qt


import operator
from functools import reduce
import numpy as np






def prod(factors):
    return reduce(operator.mul, factors, 1)



def flatten(l):
    return [item for sublist in l for item in sublist]

#tensors operators together 
def genFockOp(op,position,size,levels=2,opdim=0):
    opList=[qt.qeye(levels) for x in range(size-opdim)]
    opList[position]=op
    return qt.tensor(opList)



def numberToBase(n, b,n_qubits):
    if n == 0:
        return np.zeros(n_qubits,dtype=int)
    digits = np.zeros(n_qubits,dtype=int)
    counter=0
    while n:
        digits[counter]=int(n % b)
        n //= b
        counter+=1
    return digits[::-1]


n_qubits=7
J=1
h=0.000001



periodic=True


levels=2#
opZ=[genFockOp(qt.sigmaz(),i,n_qubits,levels) for i in range(n_qubits)]
opX=[genFockOp(qt.sigmax(),i,n_qubits,levels) for i in range(n_qubits)]
opY=[genFockOp(qt.sigmay(),i,n_qubits,levels) for i in range(n_qubits)]
opHadamard=[genFockOp(qt.qip.operations.hadamard_transform(1),i,n_qubits,levels) for i in range(n_qubits)]
opCNOT=[[0 if i==j else qt.qip.operations.cnot(n_qubits,i,j) for j in range(n_qubits)] for i in range(n_qubits)]



opId=genFockOp(qt.qeye(levels),0,n_qubits)
opZero=opId*0



base_states=np.array([numberToBase(i, 2,n_qubits) for i in range(2**n_qubits)],dtype=int)
print('base states =', base_states)
H=0

##We work in basis with ZZ and X operations
for i in range(n_qubits):
    if(i<n_qubits-1 or periodic==True):
        H+=J*opZ[i]*opZ[(i+1)%n_qubits]
    H+=-h*opX[i]

    
hilbertspace=2**n_qubits
#evals,evecs=H.eigenstates()
evals,evec=H.eigenstates(sparse=True, tol=0, maxiter=100000,eigvals=min(hilbertspace-2,15))

print("Groundstate energy",evals)




state=evec[0] ##get ground state


zero_state=qt.tensor([qt.basis(2,0) for i in range(n_qubits)]) ##all zero state
one_state=qt.tensor([qt.basis(2,1) for i in range(n_qubits)]) ##all one state



state_W=sum([opX[i]*zero_state for i in range(n_qubits)])
state_W=state_W/state_W.norm()

#print('W state =', state_W)



state_kink=sum([prod([opX[j]]+[opX[(2*i+1+j)%n_qubits] for i in range(n_qubits//2)])*zero_state for j in range(n_qubits)])
state_kink+=sum([prod([opX[j]]+[opX[(2*i+1+j)%n_qubits] for i in range(n_qubits//2)])*one_state for j in range(n_qubits)])
state_kink/=state_kink.norm()



#print("Print computational states of W-kink-state with non-zero amplitude")
probs=np.abs(state_kink.data.toarray()[:,0])**2
#print(base_states[probs>10**-12])

state_transform=qt.Qobj(state_W)
for i in range(n_qubits-1):
    #print('teeee =', i,opCNOT[i][i+1])
    state_transform=opCNOT[i][i+1]*state_transform



state_transform=opHadamard[-1]*opX[-1]*state_transform

for i in range(n_qubits//2):
    state_transform=opX[2*i]*state_transform



for i in range(n_qubits-1):
    state_transform=opCNOT[n_qubits-1][n_qubits-2-i]*state_transform
    #print('teeee 2 =',n_qubits-1,n_qubits-2-i,opCNOT[n_qubits-1][n_qubits-2-i])

print('assa =', state_transform)
print('hadamard =',opHadamard[-1])


fidelity=np.abs(state_transform.overlap(state_kink))**2
print("Fidelity kink state and transformed state from W-state",fidelity)

print("Fidelity ground state and kink state",np.abs(state.overlap(state_kink))**2)




            
            
            
   
    
