# needed libraries
import numpy as np
import cmath
import itertools
from numpy import linalg as LA
import collections
from collections import OrderedDict
import scipy.integrate as integrate
from scipy.optimize import curve_fit
from scipy import linalg
import math
from scipy.linalg import sqrtm




def PauliMatrice():
    """
     Defining the Pauli matrices

    Arguments:
        none
    Returns:
        sX, sY, sZ, sI (array complex, 2x2): pauli matrices and identity

    """

    sX = np.array([[0, 1.0], [1.0, 0]])
    sY = np.array([[0, -1.0j], [1.0j, 0]])
    sZ = np.array([[1.0, 0], [0, -1.0]])
    sI = np.array([[1.0, 0], [0, 1.0]])

    return sX,sY,sZ,sI


def kink_state_generator_AF(N,mode,which):

    # generate the list to fill out
    integer_list = []
    state_list = []

    # generate the first kink state in PBC settings 
    first_state =  ''
    for i in range(0,N):
        first_state = first_state + (str(i % 2))

    # generate the second (the same as first but globally inverted) kink state in PBC settings
    second_state = ''
    for i in range(0,N):
        second_state = second_state + (str(((i + 1) % 2) ))

    # loop overt the first state to generate all the possible kink states
    if (which == 0):
        loop_state = first_state
    if (which == 1):
        loop_state = second_state


    for j in range(0,N-1):
        new_state = str(loop_state[N-1])
        for i in range(1,N):
            new_state = new_state + str(loop_state[i-1])
        state_list.append(new_state)
        integer_list.append(int(new_state,2))

        loop_state = new_state


    if (mode == 1):
        if (which == 0):
            integer_list.append(int(first_state,2))     
            state_list.append(first_state)
        if (which == 1):
            integer_list.append(int(second_state,2))     
            state_list.append(second_state)


    return state_list,integer_list


def W_state_generator_AF(N,mode,which):

    # generate the list to fill out
    integer_list = []
    state_list = []

    # generate the first kink state in PBC settings 
    first_state =  ''
    for i in range(0,N):
        first_state = first_state + (str(i % 2))

    # generate the second (the same as first but globally inverted) kink state in PBC settings
    second_state = ''
    for i in range(0,N):
        second_state = second_state + (str(((i + 1) % 2) ))

    # loop overt the first state to generate all the possible kink states
    if (which == 0):
        loop_state = first_state
    if (which == 1):
        loop_state = second_state


    for j in range(0,N-1):
        new_state = str(loop_state[N-1])
        for i in range(1,N):
            new_state = new_state + str(loop_state[i-1])
        state_list.append(new_state)
        integer_list.append(int(new_state,2))

        loop_state = new_state


    if (mode == 1):
        if (which == 0):
            integer_list.append(int(first_state,2))     
            state_list.append(first_state)
        if (which == 1):
            integer_list.append(int(second_state,2))     
            state_list.append(second_state)


    return state_list,integer_list



def generate_kink_state(N,mode):

    kink_list, kink_int_list = kink_state_generator_AF(N,mode,0)
    kink_list_size = len(kink_list)
    zero    = np.array([1.0,0.0])
    one     = np.array([0.0,1.0])

    # generate the total state
    totstate = np.zeros(int(2**N))

    for jj in range(0,kink_list_size):
        
        # last element
        if (int(kink_list[jj][N-1]) == 0):
            last = zero.copy()
        elif(int(kink_list[jj][N-1]) == 1):
            last = one.copy()
        else:
            print('ERROR')  

        for j in range(2,N+1):

            if (int(kink_list[jj][N-j]) == 0):
                tampon = zero.copy()
            elif(int(kink_list[jj][N-j]) == 1):
                tampon = one.copy()
            else:
                print('ERROR')  

            last = np.kron(tampon, last)


        totstate = totstate + (1.0/np.sqrt(N))*last

    return totstate


def generate_kink_stateXY(N,mode,p):

    kink_list, kink_int_list = kink_state_generator_AF(N,mode,0)
    kink_list_size = len(kink_list)
    zero    = np.array([1.0,0.0])
    one     = np.array([0.0,1.0])

    #print(kink_list)
    # generate the total state
    totstate1 = np.zeros(int(2**N))

    for jj in range(0,kink_list_size):
        
        # last element
        if (int(kink_list[jj][N-1]) == 0):
            last = zero.copy()
        elif(int(kink_list[jj][N-1]) == 1):
            last = one.copy()
        else:
            print('ERROR')  

        for j in range(2,N+1):

            if (int(kink_list[jj][N-j]) == 0):
                tampon = zero.copy()
            elif(int(kink_list[jj][N-j]) == 1):
                tampon = one.copy()
            else:
                print('ERROR')  

            last = np.kron(tampon, last)


        totstate1 = totstate1 + (1.0/np.sqrt(N))*np.exp(1j*(jj+1)*p)*last


    kink_list, kink_int_list = kink_state_generator_AF(N,mode,1)
    kink_list_size = len(kink_list)

    #print(kink_list)


    totstate2 = np.zeros(int(2**N))

    for jj in range(0,kink_list_size):
        
        # last element
        if (int(kink_list[jj][N-1]) == 0):
            last = zero.copy()
        elif(int(kink_list[jj][N-1]) == 1):
            last = one.copy()
        else:
            print('ERROR')  

        for j in range(2,N+1):

            if (int(kink_list[jj][N-j]) == 0):
                tampon = zero.copy()
            elif(int(kink_list[jj][N-j]) == 1):
                tampon = one.copy()
            else:
                print('ERROR')  

            last = np.kron(tampon, last)


        totstate2 = totstate2 + (1.0/np.sqrt(N))*np.exp(-1j*(jj+1)*p)*last


    totstate = (1.0/np.sqrt(2.0))*(totstate1 + totstate2)



    return totstate


def kink_state_generator_W(N):

    # generate the list to fill out
    integer_list = []
    state_list = []

    # generate the first kink state in PBC settings 
    first_state =  ''
    for i in range(0,N):
        first_state = first_state + str(0)



    for i in range(0,N):
        loop_state = list(first_state)
        loop_state[i] = '1' 
        loop_state = ''.join(loop_state)
        state_list.append(loop_state)


    return state_list


def generate_kink_stateW(N):

    kink_list = kink_state_generator_W(N)
    kink_list_size = len(kink_list)
    
    zero    = np.array([1.0,0.0])
    one     = np.array([0.0,1.0])

    # generate the total state
    totstate = np.zeros(int(2**N))

    for jj in range(0,kink_list_size):
        
        # last element
        if (int(kink_list[jj][N-1]) == 0):
            last = zero.copy()
        elif(int(kink_list[jj][N-1]) == 1):
            last = one.copy()
        else:
            print('ERROR')  

        for j in range(2,N+1):

            if (int(kink_list[jj][N-j]) == 0):
                tampon = zero.copy()
            elif(int(kink_list[jj][N-j]) == 1):
                tampon = one.copy()
            else:
                print('ERROR')  

            last = np.kron(tampon, last)


        totstate = totstate + (1.0/np.sqrt(N))*last


    return totstate



def generate_normal_state(N):

    zero    = np.array([1.0,0.0])
    one     = np.array([0.0,1.0])

    # generate the total state
    magic0 = one
    magic  = one

    for i in range(1,N):
        magic = np.kron(magic0,magic)



    return magic


def tfim_LocalHamiltonian_new(lambdaa: float):
    """ 
    The definition of the local Trasverse Ising Hamiltonian - Fabio/Marco
    
    Arguments:
        lambdaa(float): value of the coupling for this type of Hamiltonian
    Returns:
        hloc (2x2 array float): local Hamiltonian that we apply to two neighboring sites in order to diagonaliza the Hamiltonian


     """ 

    sX, sY, sZ, sI = PauliMatrice()

    hloc = (lambdaa*np.kron(sX, sX) - 1.0*0.5*(np.kron(sZ,sI) + np.kron(sI,sZ)) ).reshape(2, 2, 2, 2)

    return hloc 


def tfim_LocalHamiltonian_proper(h: float):
    """ 
    The definition of the local Trasverse Ising Hamiltonian - Fabio/Marco
    
    Arguments:
        lambdaa(float): value of the coupling for this type of Hamiltonian
    Returns:
        hloc (2x2 array float): local Hamiltonian that we apply to two neighboring sites in order to diagonaliza the Hamiltonian


     """ 

    sX, sY, sZ, sI = PauliMatrice()

    hloc = (np.kron(sX, sX) - h*0.5*(np.kron(sZ,sI) + np.kron(sI,sZ)) ).reshape(2, 2, 2, 2)

    return hloc 



def tfim_LocalHamiltonian_properJ(h: float):
    """ 
    The definition of the local Trasverse Ising Hamiltonian - Fabio/Marco
    
    Arguments:
        lambdaa(float): value of the coupling for this type of Hamiltonian
    Returns:
        hloc (2x2 array float): local Hamiltonian that we apply to two neighboring sites in order to diagonaliza the Hamiltonian


     """ 

    sX, sY, sZ, sI = PauliMatrice()

    hloc = (-np.kron(sX, sX) - h*0.5*(np.kron(sZ,sI) + np.kron(sI,sZ)) ).reshape(2, 2, 2, 2)

    return hloc 



def tfim_LocalHamiltonian_test(lambdaa: float):
    """ 
    The definition of the local Trasverse Ising Hamiltonian - Fabio/Marco
    
    Arguments:
        lambdaa(float): value of the coupling for this type of Hamiltonian
    Returns:
        hloc (2x2 array float): local Hamiltonian that we apply to two neighboring sites in order to diagonaliza the Hamiltonian


     """ 

    sX, sY, sZ, sI = PauliMatrice()

    hloc = (np.kron(sX, sX) - lambdaa*0.5*(np.kron(sZ,sI) + np.kron(sI,sZ)) ).reshape(2, 2, 2, 2)

    return hloc 




def doApplyHamTENSOR(psi: np.ndarray, h: np.ndarray, N: int, usePBC: bool):
    """ 
    Basic succesive application of the local 4 leg local Hamiltonian to the psi tensor. 
    This routine is designed to work with the LinearOperator function and the Lanczos type algorithms.
    
    Arguments:
        psi:        vector of length d**N describing the quantum state.
        h:          array of ndim=4 describing the nearest neighbor coupling (local Hamiltonian of dimension (d^2, d^2)).
        N:          the number of lattice sites.
        usePBC:     sets whether to include periodic boundary term.
    Returns:
        psiOut:     state psi after application of the Hamiltonian.

        Logic of the application is explained somewhere in the notes. 
        First, second and last application of the local Hamiltonian are written explicelty. The remaining steps are generic and the same algorithm is applied.      
    """

    # this is the local dimension 
    d = 2

    # reshape the vector into a tensor with N legs each with dimension d
    reshape_array_default     = np.full(N,d)
    psi = psi.reshape(reshape_array_default)

    # First application of the local Hamiltonian to the tensor quantum state for legs 1 and 2 (k = 0)
    psiOut = (h.reshape(4,4) @ psi.reshape(4,2**(N-2))).reshape(reshape_array_default).reshape(2**N)


    # Second application of the local Hamiltonian to the tensor quantum state for legs 2 and 3 (k = 1)
    # generate the lists of transposition (first transpose different from the second one)
    transpose_array_default1    = np.arange(N)
    transpose_array_default2    = transpose_array_default1.copy()

    transpose_array_default1[0] = 1
    transpose_array_default1[1] = 2
    transpose_array_default1[2] = 0 

    transpose_array_default2[0] = 2
    transpose_array_default2[1] = 0
    transpose_array_default2[2] = 1 

    psiOut += (h.reshape(4,4) @ psi.transpose(transpose_array_default1).reshape(4,2**(N-2))).reshape(reshape_array_default).transpose(transpose_array_default2).reshape(2**N)


    # Remaining applications of the local Hamiltonian  (k >= 2)
    for k in range(2,N-1):

        transpose_array_default         = np.arange(N)
        transpose_array_default[0]      = k
        transpose_array_default[1]      = k+1
        transpose_array_default[k]      = 0 
        transpose_array_default[k+1]    = 1

        psiOut += (h.reshape(4,4) @ psi.transpose(transpose_array_default).reshape(4,2**(N-2))).reshape(reshape_array_default).transpose(transpose_array_default).reshape(2**N)

        
    # PBC conditions application of the local hamiltonian on the end tails of the psi tensor
    if usePBC:

        # generate the lists of transposition (first transpose different from the second one)
        transpose_array_default1  = np.arange(N)
        transpose_array_default2  = transpose_array_default1.copy()

        transpose_array_default1[0]    = N - 1
        transpose_array_default1[1]    = 0
        transpose_array_default1[N-1]  = 1

        transpose_array_default2[0]    = 1
        transpose_array_default2[1]    = N-1
        transpose_array_default2[N-1]  = 0

        psiOut += (h.reshape(4,4) @ psi.transpose(transpose_array_default1).reshape(4,2**(N-2))).reshape(reshape_array_default).transpose(transpose_array_default2).reshape(2**N)


    return psiOut    




def local_magnetization_single_point(sigma,j,psi,N,d):
    """ function to evaluate single point correlator < psi | sigma_{j}^{x,y,z} | psi > - magnetization """ 

    initial_psi = psi.copy()

    # reshape into a tensor with N legs, each leg with dimension d
    reshape_array_default     = np.full(N,d)    
    psi = psi.reshape(reshape_array_default)

    # transpose the tensor with N legs to match the position of the local operator
    transpose_array_default        = np.arange(N)
    transpose_array_default[0]     = j-1
    transpose_array_default[j-1]   = 0

    psi = (sigma @ psi.transpose(transpose_array_default).reshape(d,2**(N-1))).reshape(reshape_array_default).transpose(transpose_array_default).reshape(2**N)

    mag = (np.transpose(np.conjugate(initial_psi))) @ psi

    return mag  


# generate the list momentum indices
def O_array(N):

    # definitions
    prefactor = np.pi/N
    R = int(N/2)

    array = []
    for i in range(0,R):

        value = N - 2*(i + 1)
        array.append(prefactor*value)
        array.append(-prefactor*value)

    array.append(np.pi)
    array = np.sort(array)

    return array


# generate the list momentum indices, subset
# mode = 0, full list
# mode = 1, full list minus the pi/N term
def O_array_plus(N):

    # definitions
    prefactor = np.pi/N
    R = int(N/2)

    array = []
    for i in range(0,R):

        value = N - 2*(i + 1)
        array.append(prefactor*value)


    array = np.sort(array)

    return array


# generate the list momentum indices
def E_array(N):

    # definitions
    prefactor = np.pi/N
    R = int(N/2)

    array = []
    for i in range(0,R):

        value = N - (2*(i + 1) - 1 )
        #print(N,i,value, 2*(i + 1) - 1 )
        array.append(prefactor*value)
        array.append(-prefactor*value)

    array.append(0.0)
    array = np.sort(array)

    return array


# generate the list momentum indices, subset
# mode = 0, full list
# mode = 1, full list minus the pi/N term
def E_array_plus(N):

    # definitions
    prefactor = np.pi/N
    R = int(N/2)

    array = []
    for i in range(0,R):

        value = N - (2*(i + 1) - 1 )
        #print(N,i,value, 2*(i + 1) - 1 )
        array.append(prefactor*value)

    array = np.sort(array)

    return array   


def XY_E_o_1(N,h,gamma):

    lista = O_array_plus(N)

    summ = 0.0

    for i in range(0,len(lista)):
        k = lista[0]
        value = np.sqrt((np.cos(k) + h)**2.0 + (gamma**2.0)*(np.sin(k)**2.0))

        summ = summ + value

    energy = 1.0 - h - 2.0*summ

    return energy

def XY_E_e_1(N,h,gamma):

    lista = E_array_plus(N)

    summ = 0.0

    for i in range(0,len(lista)):
        k = lista[0]


        value = np.sqrt((np.cos(k) + h)**2.0 + (gamma**2.0)*(np.sin(k)**2.0))

        summ = summ + value

    energy = 1.0 + h - 2.0*summ

    return energy



def XY_E_o_2(N,h,gamma):

    lista = O_array_plus(N)

    summ = 0.0

    for i in range(0,len(lista)):
        k = lista[0]

        if (abs((np.pi)/N - k) < 10e-14):
            print('found it !')
        else:        
            value = np.sqrt((np.cos(k) + h)**2.0 + (gamma**2.0)*(np.sin(k)**2.0))

        summ = summ + value

    energy = h - 1.0 - 2.0*summ

    return energy


def XY_E_e_2(N,h,gamma):

    lista = O_array_plus(N)

    summ = 0.0

    for i in range(0,len(lista)):
        k = lista[i]

        if (abs((np.pi*(N - 1.0))/N - k) < 10e-14):
            print('found it !')
        else:
            value = np.sqrt((np.cos(k) + h)**2.0 + (gamma**2.0)*(np.sin(k)**2.0))

        summ = summ + value

    energy = -(h + 1.0) - 2.0*summ

    return energy




def epsilon_k(k,h,gamma):

    value = (np.cos(k) + h)/(np.sqrt( (np.cos(k) + h)**2.0 + (gamma*np.sin(k))**2.0 ))

    return value

def delta_k(k,h,gamma):

    value = (gamma*np.sin(k))/(np.sqrt( (np.cos(k) + h)**2.0 + (gamma*np.sin(k))**2.0 ))

    return value


def g_0_10(N,r,h,gamma):

    summ = 0.0
    lista = O_array(N)

    for i in range(0,len(lista)):
        k = lista[i]
        value1 = epsilon_k(k,h,gamma)
        value2 = delta_k(k,h,gamma)

        summ = summ + value1*np.cos(r*k) + value2*np.sin(r*k)

    value = summ/N

    return value


def g_1_10(N,r,h,gamma):

    if (h > 0.0):
        value = 0.0

    if (h < 1.0):
        value = (2.0*((-1.0)**(r)))/N

    return value





def local_magnetization_general_point_general(sigma,jlist,psi,N,d):
    """ # function to evaluate two point spin correlator < psi | sigma_{j1}^{x,y,z} sigma_{j2}^{x,y,z} ... sigma_{jN}^{x,y,z}| psi > - ONLY MADE TO WORK WITH PBC IMPOSED ON THE PROBLEM (see later mod function % which loops around psi) """

    initial_psi = psi.copy()

    # reshape into a tensor with N legs, each leg with dimension d
    reshape_array_default         = np.full(N,d)    
    psi = psi.reshape(reshape_array_default)

    # general loop
    for ii in range(0,len(jlist)):
        # transpose the tensor with N legs to match the position of the local operator
        transpose_array_default       = np.arange(N)
        transpose_array_default[0]    = jlist[ii]-1
        transpose_array_default[jlist[ii]-1]  = 0

        # first operator application
        psi = (sigma[ii] @ psi.transpose(transpose_array_default).reshape(d,2**(N-1))).reshape(reshape_array_default).transpose(transpose_array_default)

    psi = psi.reshape(2**N)

    mag = np.transpose(initial_psi) @ psi

    return mag  



def local_magnetization_two_point_general(sigma2,sigma1,j,t,psi,N,d):
    """ # function to evaluate two point spin correlator < psi | sigma_{j + t}^{x,y,z} sigma_{j}^{x,y,z} | psi > - ONLY MADE TO WORK WITH PBC IMPOSED ON THE PROBLEM (see later mod function % which loops around psi) """

    initial_psi = psi.copy()

    # reshape into a tensor with N legs, each leg with dimension d
    reshape_array_default         = np.full(N,d)    
    psi = psi.reshape(reshape_array_default)

    # transpose the tensor with N legs to match the position of the local operator
    transpose_array_default       = np.arange(N)
    transpose_array_default[0]    = j-1
    transpose_array_default[j-1]  = 0

    # first operator application
    psi = (sigma1 @ psi.transpose(transpose_array_default).reshape(d,2**(N-1))).reshape(reshape_array_default).transpose(transpose_array_default)

    # transpose the tensor with N legs to match the position of the local operator
    transpose_array_default       = np.arange(N)
    transpose_array_default[0]    = t-1               
    transpose_array_default[t-1]  = 0


    # second operator application
    psi = (sigma2 @ psi.transpose(transpose_array_default).reshape(d,2**(N-1))).reshape(reshape_array_default).transpose(transpose_array_default).reshape(2**N)

    mag = np.transpose(initial_psi) @ psi

    return mag  


def local_magnetization_two_point(sigma,j,t,psi,N,d):
    """ # function to evaluate two point spin correlator < psi | sigma_{j + t}^{x,y,z} sigma_{j}^{x,y,z} | psi > - ONLY MADE TO WORK WITH PBC IMPOSED ON THE PROBLEM (see later mod function % which loops around psi) """

    initial_psi = psi.copy()

    # reshape into a tensor with N legs, each leg with dimension d
    reshape_array_default         = np.full(N,d)    
    psi = psi.reshape(reshape_array_default)

    # transpose the tensor with N legs to match the position of the local operator
    transpose_array_default       = np.arange(N)
    transpose_array_default[0]    = j-1
    transpose_array_default[j-1]  = 0

    # first operator application
    psi = (sigma @ psi.transpose(transpose_array_default).reshape(d,2**(N-1))).reshape(reshape_array_default).transpose(transpose_array_default)

    # transpose the tensor with N legs to match the position of the local operator
    transpose_array_default       = np.arange(N)
    transpose_array_default[0]    = (j+t-1) % N               
    transpose_array_default[(j+t-1) % N]  = 0


    # second operator application
    psi = (sigma @ psi.transpose(transpose_array_default).reshape(d,2**(N-1))).reshape(reshape_array_default).transpose(transpose_array_default).reshape(2**N)

    mag = np.transpose(initial_psi) @ psi

    return mag  
    



def generate_kink_stateUPDOWN(N,mode):

    kink_list, kink_int_list = kink_state_generator_AF(N,mode,0)
    kink_list_size = len(kink_list)
    zero    = (1.0/np.sqrt(2.0))*(np.array([1.0,0.0]) + np.array([0.0,1.0]))
    one     = (1.0/np.sqrt(2.0))*(np.array([1.0,0.0]) - np.array([0.0,1.0]))

    # generate the total state
    totstate = np.zeros(int(2**N))

    for jj in range(0,kink_list_size):
        
        # last element
        if (int(kink_list[jj][N-1]) == 0):
            last = zero.copy()
        elif(int(kink_list[jj][N-1]) == 1):
            last = one.copy()
        else:
            print('ERROR')  

        for j in range(2,N+1):

            if (int(kink_list[jj][N-j]) == 0):
                tampon = zero.copy()
            elif(int(kink_list[jj][N-j]) == 1):
                tampon = one.copy()
            else:
                print('ERROR')  

            last = np.kron(tampon, last)


        totstate = totstate + (1.0/np.sqrt(2*N))*last





    return totstate    


def generate_kink_stateUPDOWNONE(N,jj):

    kink_list, kink_int_list = kink_state_generator_AF(N,1,0)
    kink_list_size = len(kink_list)
    zero    = (1.0/np.sqrt(2.0))*(np.array([1.0,0.0]) + np.array([0.0,1.0]))
    one     = (1.0/np.sqrt(2.0))*(np.array([1.0,0.0]) - np.array([0.0,1.0]))

    # generate the total state
    totstate = np.zeros(int(2**N))

    #print(kink_list,kink_list[jj]) 
    
    # last element
    if (int(kink_list[jj][N-1]) == 0):
        last = zero.copy()
        #print('zero')
    elif(int(kink_list[jj][N-1]) == 1):
        last = one.copy()
        #print('one')
    else:
        print('ERROR')  

    for j in range(2,N+1):

        if (int(kink_list[jj][N-j]) == 0):
            tampon = zero.copy()
            #print('zero')
        elif(int(kink_list[jj][N-j]) == 1):
            tampon = one.copy()
            #print('one')
        else:
            print('ERROR')  

        last = np.kron(tampon, last)


    totstate = totstate + last




    return totstate      



def local_magnetization_single_point(sigma,j,psi,N,d):
    """ function to evaluate single point correlator < psi | sigma_{j}^{x,y,z} | psi > - magnetization """ 

    initial_psi = psi.copy()

    # reshape into a tensor with N legs, each leg with dimension d
    reshape_array_default     = np.full(N,d)    
    psi = psi.reshape(reshape_array_default)

    # transpose the tensor with N legs to match the position of the local operator
    transpose_array_default        = np.arange(N)
    transpose_array_default[0]     = j-1
    transpose_array_default[j-1]   = 0

    psi = (sigma @ psi.transpose(transpose_array_default).reshape(d,2**(N-1))).reshape(reshape_array_default).transpose(transpose_array_default).reshape(2**N)

    mag = (np.transpose(np.conjugate(initial_psi))) @ psi

    return mag 



def doApplyHamTENSOR_FULL_DISORDER(psi: np.ndarray,h: np.ndarray, N: int, usePBC: bool):
    """ Basic succesive application of the local 4 leg local Hamiltonian to the psi tensor. 
        This routine is designed to work with the LinearOperator function and the Lanczos type algorithms.
        Args:
            psi:        vector of length d**N describing the quantum state.
            N:          the number of lattice sites.
            usePBC:     sets whether to include periodic boundary term.
            position:   position of the defect
        Returns:
            np.ndarray: state psi after application of the Hamiltonian.

        Logic of the application is explained somewhere in the notes. 
        First, second and last application of the local Hamiltonian are written explicelty. The remaining steps are generic and the same algorithm is applied.      
    """
    sX, sY, sZ, sI = PauliMatrice()



    # this is the local dimension 
    d = 2

    # reshape the vector into a tensor with N legs each with dimension d
    reshape_array_default     = np.full(N,d)
    psi = psi.reshape(reshape_array_default)

    # First application of the local Hamiltonian to the tensor quantum state for legs 1 and 2 (k = 0)
    psiOut = (h[0].reshape(4,4) @ psi.reshape(4,2**(N-2))).reshape(reshape_array_default).reshape(2**N)


    # Second application of the local Hamiltonian to the tensor quantum state for legs 2 and 3 (k = 1)
    # generate the lists of transposition (first transpose different from the second one)
    transpose_array_default1    = np.arange(N)
    transpose_array_default2    = transpose_array_default1.copy()

    transpose_array_default1[0] = 1
    transpose_array_default1[1] = 2
    transpose_array_default1[2] = 0 

    transpose_array_default2[0] = 2
    transpose_array_default2[1] = 0
    transpose_array_default2[2] = 1 


    psiOut += (h[1].reshape(4,4) @ psi.transpose(transpose_array_default1).reshape(4,2**(N-2))).reshape(reshape_array_default).transpose(transpose_array_default2).reshape(2**N)


    # Remaining applications of the local Hamiltonian  (k >= 2)
    for k in range(2,N-1):

        transpose_array_default         = np.arange(N)
        transpose_array_default[0]      = k
        transpose_array_default[1]      = k+1
        transpose_array_default[k]      = 0 
        transpose_array_default[k+1]    = 1

        psiOut += (h[k].reshape(4,4) @ psi.transpose(transpose_array_default).reshape(4,2**(N-2))).reshape(reshape_array_default).transpose(transpose_array_default).reshape(2**N)

        
    # PBC conditions application of the local hamiltonian on the end tails of the psi tensor
    if usePBC:

        # generate the lists of transposition (first transpose different from the second one)
        transpose_array_default1  = np.arange(N)
        transpose_array_default2  = transpose_array_default1.copy()

        transpose_array_default1[0]    = N - 1
        transpose_array_default1[1]    = 0
        transpose_array_default1[N-1]  = 1

        transpose_array_default2[0]    = 1
        transpose_array_default2[1]    = N-1
        transpose_array_default2[N-1]  = 0

  
        psiOut += (h[N-1].reshape(4,4) @ psi.transpose(transpose_array_default1).reshape(4,2**(N-2))).reshape(reshape_array_default).transpose(transpose_array_default2).reshape(2**N)


    return psiOut



def tfim_LocalHamiltonian_DISORDER(J_disorder_array,h_disorder_array):

    """ Trasverse Ising Hamiltonian with random couplings and on-site magnetic field """ 

    sX, sY, sZ, sI = PauliMatrice()

    hloc = []
    for i in range(0,len(J_disorder_array)):
        #hdisorder = (J_disorder_array[i]*np.kron(sX, sX) - 0.5*h_disorder_array[i]*(np.kron(sI,sZ) + np.kron(sZ,sI)) ).reshape(2, 2, 2, 2)
        hdisorder = (J_disorder_array[i]*np.kron(sX, sX) - h_disorder_array[i]*(np.kron(sZ,sI) ) ).reshape(2, 2, 2, 2)

        hloc.append(hdisorder)
    
    return hloc  




def partition_vonNeuman(Nsites,list, psi):

    rho_reduced = PartialTraceGeneralTensor(Nsites,list,psi) 

    eigenvalues,eigenvectors = LA.eigh(rho_reduced)
    idx = eigenvalues.argsort()[::-1]   
    eigenvalues = eigenvalues[idx]

    # compute the von Neumann entropy
    entropy = 0.0
    epsi = 10**(-15)

    for ii in range(0,len(eigenvalues)):
        if (eigenvalues[ii] > epsi):
            entropy += eigenvalues[ii]*np.log(eigenvalues[ii])

    return -entropy



def PartialTraceGeneralTensor(N,index_list,A):
    """ Function that computes the partial trace over index_list indices (the index list needs to be ordered from smaller to bigger index)"""

    # reshape the input vectors into tensors (here we exploit the fact that psi* is just the complex conjugate of psi )
    reshape_array_default = np.full(N,2)    
    A_initial = A.reshape(reshape_array_default)

    # generate initial transpose indices vector (we apply permutations and operatorion so transposition is correctly performed )
    list_A = np.arange(N)
    list_B = np.arange(N)

    # this changing the indeces by one is because of python stuff (the numbering starts from zero and not 1)
    index_list = np.array(index_list) - 1

    ##### generating the first transpose rule for A ###

    ## initial step of moving the indices to the left
    for zz in range(0,len(index_list)):
        list_A[zz] = index_list[zz]

    ## figure out what are the missing indices that happen because of overwritting in loop above
    list_A_no_dupl = list(OrderedDict.fromkeys(list_A))
    missing_indices = np.delete(np.arange(N), list_A_no_dupl)

    ## now replace the doubled indices with indices in the missing_indices array
    counter = 0
    for zz in range(len(index_list),len(list_A)):
        for tt in range(0,len(index_list)):
            if (list_A[zz] == index_list[tt]):
                list_A[zz] = missing_indices[counter]
                counter += 1
    

    ##### generating the first transpose rule for B ###

    ## initial step of moving the indices to the right
    for zz in range(0,len(index_list)):
        list_B[len(list_B) - zz - 1] = index_list[len(index_list) - zz - 1]

    ## figure out what are the missing indices that happen because of overwritting in loop above
    list_B_no_dupl = list(OrderedDict.fromkeys(list_B))
    missing_indices = np.delete(np.arange(N), list_B_no_dupl)

    ## now replace the doubled indices with indices in the missing_indices array
    counter = 0
    for zz in range(0,len(list_B)-len(index_list)):
        for tt in range(0,len(index_list)):
            if (list_B[zz] == index_list[tt]):
                list_B[zz] = missing_indices[counter]
                counter += 1


    ##### generating the second transpose rule for A ###

    list_A_cut = list_A[len(index_list):]
    list_A_cut_sort = np.sort(list_A_cut)

    list_B_cut = list_B[:-len(index_list)]
    list_B_cut_sort = np.sort(list_B_cut)

    transpose2_A = np.append(index_list,list_A_cut_sort)
    transpose2_B = np.append(list_B_cut_sort,index_list)

    ############### MAIN OPERATION AFTER ALL PREPARATION HAS BEEN PERFORMED ::: TRANSPOSITION ON A and B
    A = A_initial.transpose(transpose2_A).reshape(2**len(index_list),2**(N - len(index_list)))
    B = A_initial.transpose(transpose2_B).reshape(2**(N - len(index_list)),2**len(index_list))

    # FINAL MULTIPLICATION
    out = (A @ np.conjugate(B))

    return out    




def Renyi2DIRECT(N,A,psi):
    """

    compute renyi2 for just the thing we are looking for

    """

    rho_reduced = PartialTraceGeneralTensor(N,A,psi) 

    entropy = -np.log((rho_reduced * rho_reduced.T).sum())                              #-np.log(np.trace(rho_reduced @ rho_reduced)) #-np.log(np.sum(np.square(rho_reduced)))

    return entropy     




def generate_kink_stateTOT(N,mode):

    kink_list, kink_int_list = kink_state_generator_AF(N,mode,0)
    kink_list_size = len(kink_list)
    zero    = np.array([1.0,0.0])
    one     = np.array([0.0,1.0])

    # generate the total state
    totstate = np.zeros(int(2**N))

    for jj in range(0,kink_list_size):
        
        # last element
        if (int(kink_list[jj][N-1]) == 0):
            last = zero.copy()
        elif(int(kink_list[jj][N-1]) == 1):
            last = one.copy()
        else:
            print('ERROR')  

        for j in range(2,N+1):

            if (int(kink_list[jj][N-j]) == 0):
                tampon = zero.copy()
            elif(int(kink_list[jj][N-j]) == 1):
                tampon = one.copy()
            else:
                print('ERROR')  

            last = np.kron(tampon, last)


        totstate = totstate + (1.0/np.sqrt(2.0*N))*last


    kink_list, kink_int_list = kink_state_generator_AF(N,mode,1)
    kink_list_size = len(kink_list)

    for jj in range(0,kink_list_size):
        
        # last element
        if (int(kink_list[jj][N-1]) == 0):
            last = zero.copy()
        elif(int(kink_list[jj][N-1]) == 1):
            last = one.copy()
        else:
            print('ERROR')  

        for j in range(2,N+1):

            if (int(kink_list[jj][N-j]) == 0):
                tampon = zero.copy()
            elif(int(kink_list[jj][N-j]) == 1):
                tampon = one.copy()
            else:
                print('ERROR')  

            last = np.kron(tampon, last)


        totstate = totstate + (1.0/np.sqrt(2.0*N))*last


    return totstate